@@ -7,59 +7,50 @@
 
 #include "eog-collection-item.h"
 #include "eog-image.h"
+#include "eog-canvas-pixbuf.h"
 
 struct _EogCollectionItemPrivate {
 	EogImage *image;
 
 	gboolean selected;
 
-	GnomeCanvasItem *background;
 	GnomeCanvasItem *frame;
 	GnomeCanvasItem *pixbuf_item;
 	GnomeCanvasItem *caption_item;
 	GnomeCanvasItem *selected_item;
 };
 
-#define EOG_COLLECTION_ITEM_THUMB_WIDTH  96
-#define EOG_COLLECTION_ITEM_THUMB_HEIGHT 96
-#define EOG_COLLECTION_ITEM_CAPTION_PADDING 6
+
+enum {
+	SIGNAL_SIZE_CHANGED,
+	SIGNAL_LAST
+};
+static gint eog_collection_item_signals [SIGNAL_LAST];
 
 static void eog_collection_item_destroy (GtkObject *object);
 
 GNOME_CLASS_BOILERPLATE (EogCollectionItem, eog_collection_item,
 			 GnomeCanvasGroup, GNOME_TYPE_CANVAS_GROUP);
 
-static double
-eog_collection_item_point (GnomeCanvasItem *item, double x, double y, int cx, int cy,
-			   GnomeCanvasItem **actual_item)
-{
-	double x1, y1, x2, y2;
-	double width, height;
-
-	*actual_item = item;
-
-	gnome_canvas_item_get_bounds (item, &x1, &y1, &x2, &y2);
-
-	width = x2 - x1;
-	height = y2 - y1;
-
-	return ((x >= 0) && (y >= 0) && (x <= width) && (y <= height)) ? 0.0 : 1.0;
-}
-
 static void
 eog_collection_item_class_init (EogCollectionItemClass *klass)
 {
         GObjectClass *gobject_class;
 	GtkObjectClass *object_class;
-	GnomeCanvasItemClass *item_class;
 
         gobject_class = (GObjectClass *) klass;
 	object_class = (GtkObjectClass *) klass;
-	item_class = (GnomeCanvasItemClass *) klass;
-
-	item_class->point = eog_collection_item_point;
 
 	object_class->destroy = eog_collection_item_destroy;
+
+	eog_collection_item_signals [SIGNAL_SIZE_CHANGED] = 
+		g_signal_new ("size_changed",
+			      G_TYPE_OBJECT,
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (EogCollectionItemClass, size_changed),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__VOID,
+			      G_TYPE_NONE, 0);			     
 }
 
 static void
@@ -104,10 +95,13 @@ eog_collection_item_destroy (GtkObject *object)
 }
 
 static void
-set_pixbuf (EogCollectionItem *item, GdkPixbuf *pixbuf)
+set_pixbuf (EogCollectionItem *item, GdkPixbuf *pixbuf, gboolean view_frame)
 {
 	EogCollectionItemPrivate *priv;
 	GdkPixbuf *scaled;
+	GdkPixbuf *old_pixbuf;
+	int old_width = 0;
+	int old_height = 0;
 	int image_width;
 	int image_height;
 	int image_x;
@@ -115,6 +109,14 @@ set_pixbuf (EogCollectionItem *item, GdkPixbuf *pixbuf)
 
 	priv = item->priv;
 
+	/* determine size of the old pixbuf */
+	g_object_get (G_OBJECT (priv->pixbuf_item), "pixbuf", &old_pixbuf, NULL);
+	if (old_pixbuf != NULL) {
+		old_width = gdk_pixbuf_get_width (old_pixbuf);
+		old_height = gdk_pixbuf_get_height (old_pixbuf);
+		g_object_unref (old_pixbuf);
+	}
+
 	image_width = gdk_pixbuf_get_width (pixbuf);
 	image_height = gdk_pixbuf_get_height (pixbuf);
 
@@ -158,15 +160,24 @@ set_pixbuf (EogCollectionItem *item, GdkPixbuf *pixbuf)
 			       NULL);
 	
 	gnome_canvas_item_set (priv->frame,
-			       "x1", (double) (image_x - 1),
-			       "y1", (double) (image_y - 1),
+			       "x1", (double) (image_x - EOG_COLLECTION_ITEM_FRAME_WIDTH),
+			       "y1", (double) (image_y - EOG_COLLECTION_ITEM_FRAME_WIDTH),
 			       "x2", (double) (image_x + image_width - 1),
 			       "y2", (double) (image_y + image_height - 1),
 			       NULL);
-	gnome_canvas_item_show (priv->frame);
-
+	if (view_frame) {
+		gnome_canvas_item_show (priv->frame);
+	}
+	else {
+		gnome_canvas_item_hide (priv->frame);
+	}
 
 	gdk_pixbuf_unref (scaled);
+
+	/* emit size changed signal if pixbuf size changed */
+	if (image_width != old_width || image_height != old_height) {
+		g_signal_emit (G_OBJECT (item), eog_collection_item_signals [SIGNAL_SIZE_CHANGED], 0);
+	}
 }
 
 static GdkPixbuf*
@@ -230,7 +241,7 @@ thumbnail_finished_cb (EogImage *image, gpointer data)
 
 	pixbuf = eog_image_get_pixbuf_thumbnail (priv->image);
 
-	set_pixbuf (EOG_COLLECTION_ITEM (data), pixbuf);
+	set_pixbuf (EOG_COLLECTION_ITEM (data), pixbuf, TRUE);
 
 	gdk_pixbuf_unref (pixbuf);
 }
@@ -242,9 +253,7 @@ thumbnail_failed_cb (EogImage *image, gpointer data)
 
 	pixbuf = get_failed_pixbuf ();
 
-	g_print ("pixmap failed.%s", eog_image_get_caption (image));
-
-	set_pixbuf (EOG_COLLECTION_ITEM (data), pixbuf);
+	set_pixbuf (EOG_COLLECTION_ITEM (data), pixbuf, FALSE);
 
 	gdk_pixbuf_unref (pixbuf);
 }	
@@ -259,7 +268,7 @@ image_changed_cb (EogImage *image, gpointer data)
 
 	pixbuf = eog_image_get_pixbuf_thumbnail (priv->image);
 
-	set_pixbuf (EOG_COLLECTION_ITEM (data), pixbuf);
+	set_pixbuf (EOG_COLLECTION_ITEM (data), pixbuf, TRUE);
 
 	gdk_pixbuf_unref (pixbuf);
 }
@@ -358,16 +367,6 @@ eog_collection_item_construct (EogCollectionItem *item, EogImage *image)
 	priv->image = image;
 	g_object_ref (priv->image);
 
-	/* background */
-	priv->background = gnome_canvas_item_new (GNOME_CANVAS_GROUP (item),
-						  GNOME_TYPE_CANVAS_RECT,
-						  "x1", 0.0,
-						  "y2", 0.0,
-						  "x2", (double) EOG_COLLECTION_ITEM_THUMB_WIDTH,
-						  "y2", (double) EOG_COLLECTION_ITEM_THUMB_HEIGHT,
-						  "width_pixels", 1,
-						  NULL);
-
 	/* Caption */
 	layout = gtk_widget_create_pango_layout (GTK_WIDGET (GNOME_CANVAS_ITEM (item)->canvas), NULL);
 	g_assert (layout != NULL);
@@ -379,7 +378,7 @@ eog_collection_item_construct (EogCollectionItem *item, EogImage *image)
 	caption = transform_caption (basic_caption, layout);
 
 	caption_x = EOG_COLLECTION_ITEM_THUMB_WIDTH / 2;
-	caption_y = EOG_COLLECTION_ITEM_THUMB_HEIGHT + 2 + EOG_COLLECTION_ITEM_CAPTION_PADDING;
+	caption_y = EOG_COLLECTION_ITEM_THUMB_HEIGHT + 2 * EOG_COLLECTION_ITEM_FRAME_WIDTH + EOG_COLLECTION_ITEM_SPACING;
 
 	priv->caption_item = gnome_canvas_item_new (GNOME_CANVAS_GROUP (item),
 						    GNOME_TYPE_CANVAS_TEXT,
@@ -397,13 +396,13 @@ eog_collection_item_construct (EogCollectionItem *item, EogImage *image)
 	gnome_canvas_item_get_bounds (priv->caption_item, &x1, &y1, &x2, &y2); 
 	priv->selected_item = gnome_canvas_item_new (GNOME_CANVAS_GROUP (item),
 						     GNOME_TYPE_CANVAS_RECT,
-						     "x1", (double) (x1 - 2.0),
-						     "y1", (double) (y1 - 2.0),
-						     "x2", (double) (x2 + 2.0),
-						     "y2", (double) (y2 + 2.0),
+						     "x1", (double) (x1 - EOG_COLLECTION_ITEM_CAPTION_PADDING),
+						     "y1", (double) (y1 - EOG_COLLECTION_ITEM_CAPTION_PADDING),
+						     "x2", (double) (x2 + EOG_COLLECTION_ITEM_CAPTION_PADDING),
+						     "y2", (double) (y2 + EOG_COLLECTION_ITEM_CAPTION_PADDING),
 						     "fill_color", "LightSteelBlue2",
 						     "outline_color", "Blue",
-						     "width_pixels", 2,
+						     "width_pixels",  EOG_COLLECTION_ITEM_CAPTION_FRAME_WIDTH,
 						     "outline_stipple", get_stipple_bitmap (),
 						     NULL);
 	gnome_canvas_item_lower (priv->selected_item, 1);
@@ -411,13 +410,13 @@ eog_collection_item_construct (EogCollectionItem *item, EogImage *image)
 
 	/* Image */
 	priv->pixbuf_item = gnome_canvas_item_new (GNOME_CANVAS_GROUP (item),
-						   GNOME_TYPE_CANVAS_PIXBUF,
+						   EOG_TYPE_CANVAS_PIXBUF,
 						   NULL);
 
 	/* Image Frame */
 	priv->frame = gnome_canvas_item_new (GNOME_CANVAS_GROUP (item),
 					     GNOME_TYPE_CANVAS_RECT,
-					     "width_pixels", 1,
+					     "width_pixels", EOG_COLLECTION_ITEM_FRAME_WIDTH,
 					     "outline_color", "Black",
 					     NULL);
 	gnome_canvas_item_hide (priv->frame);
@@ -427,7 +426,7 @@ eog_collection_item_construct (EogCollectionItem *item, EogImage *image)
 	g_signal_connect (image, "image_changed", G_CALLBACK (image_changed_cb), item);
 
 	pixbuf = get_busy_pixbuf ();
-	set_pixbuf (item, pixbuf);
+	set_pixbuf (item, pixbuf, FALSE);
 	gdk_pixbuf_unref (pixbuf);
 
 	gnome_canvas_item_request_update (GNOME_CANVAS_ITEM (item));
@@ -508,20 +507,31 @@ eog_collection_item_get_image (EogCollectionItem *item)
 	return priv->image;
 }
 
+/* eog_collection_item_get_size:
+ *
+ * Returns the width and the height of the image and the caption seperately. This is
+ * used by EogWrapList to align the items at the image bottom
+ * line. caption_height includes the padding between image and caption
+ * so that image_height + caption_height gives the total height of the item.
+ */
 void 
-eog_collection_item_get_size (EogCollectionItem *item, int *width, int *height)
+eog_collection_item_get_size (EogCollectionItem *item, int *width, int *image_height, int *caption_height)
 {
-	int caption_height;
 	double x1, x2, y1, y2;
+	int img_width;
+	int cap_width;
 
+	/* FIXME: maybe we should cache the values here, because they get
+	 * calculated at least 2 times for every item during item-rearrangement of 
+	 * the wrap list.
+	 */
 	gnome_canvas_item_get_bounds (item->priv->selected_item, &x1, &y1, &x2, &y2); 
+	*caption_height = (int) EOG_COLLECTION_ITEM_CAPTION_PADDING + y2 - y1;
+	cap_width = (int) x2 - x1;
+	
+	gnome_canvas_item_get_bounds (item->priv->frame, &x1, &y1, &x2, &y2); 
+	*image_height = (int) y2 - y1;
+	img_width = (int) x2 - x1;
 
-	caption_height = y2 - y1;
-
-	*height = 
-		EOG_COLLECTION_ITEM_THUMB_HEIGHT + 
-		EOG_COLLECTION_ITEM_CAPTION_PADDING + 
-		caption_height;
-
-	*width = EOG_COLLECTION_ITEM_THUMB_WIDTH + 4;
+	*width = MAX (cap_width, img_width);
 }
