@@ -20,9 +20,12 @@
  */
 #include <config.h>
 #include <math.h>
+#include <string.h>
 #include <glib/gmain.h>
+#include <glib/gmacros.h>
 #include <libgnome/gnome-macros.h>
 #include <gdk/gdkkeysyms.h>
+#include <gtk/gtkdnd.h>
 
 #include "eog-wrap-list.h"
 #include "eog-collection-marshal.h"
@@ -30,7 +33,22 @@
 #include "eog-collection-item.h"
 
 #define COLLECTION_DEBUG 0
-
+
+#define EOG_WRAP_LIST_BORDER  12
+static const int MAX_ITEM_WIDTH = EOG_COLLECTION_ITEM_MAX_WIDTH;
+
+#define EOG_DND_URI_LIST_TYPE 	         "text/uri-list"
+
+
+
+enum {
+	EOG_DND_URI_LIST
+};
+
+static GtkTargetEntry drag_types [] = {
+	{ EOG_DND_URI_LIST_TYPE,   0, EOG_DND_URI_LIST }
+};
+
 
 /* Used to hold signal handler IDs for models */
 typedef enum {
@@ -50,10 +68,10 @@ enum {
 static guint eog_wrap_list_signals [LAST_SIGNAL];
 
 enum {
-	GLOBAL_SIZE_CHANGED,
+	GLOBAL_WIDGET_SIZE_CHANGED,
 	GLOBAL_SPACING_CHANGED,
-	GLOBAL_LAYOUT_MODE_CHANGED,
-	GLOBAL_HINT_LAST
+	ITEM_SIZE_CHANGED,
+	UPDATE_HINT_LAST
 };
 
 typedef struct {
@@ -95,9 +113,6 @@ struct _EogWrapListPrivate {
 	/* Number of items */
 	guint n_items; 
 
-	/* Layout mode to use for row/col calculating. */
-	EogLayoutMode lm;
-
 	/* Number of rows. */
 	guint n_rows;
 
@@ -108,7 +123,7 @@ struct _EogWrapListPrivate {
 	gint idle_handler_id;
 
 	/* Update hints */
-	gboolean global_update_hints [GLOBAL_HINT_LAST];
+	gboolean global_update_hints [UPDATE_HINT_LAST];
 
 	/* last id of thumbnail the user clicked */
 	GnomeCanvasItem *last_item_clicked;
@@ -121,6 +136,11 @@ static void eog_wrap_list_finalize (GObject *object);
 
 static void eog_wrap_list_size_allocate (GtkWidget *widget, GtkAllocation *allocation);
 static gboolean eog_wrap_list_key_press_cb (GtkWidget *widget, GdkEventKey *event);
+static void eog_wrap_list_drag_data_get_cb (GtkWidget *widget,
+					    GdkDragContext *drag_context,
+					    GtkSelectionData *data,
+					    guint info,
+					    guint time);
 
 static void request_update (EogWrapList *wlist);
 static gboolean do_update (EogWrapList *wlist);
@@ -146,6 +166,7 @@ eog_wrap_list_class_init (EogWrapListClass *class)
 
 	widget_class->size_allocate = eog_wrap_list_size_allocate;
 	widget_class->key_press_event = eog_wrap_list_key_press_cb;
+	widget_class->drag_data_get = eog_wrap_list_drag_data_get_cb;
 
 	eog_wrap_list_signals [SELECTION_CHANGED] = 
 		g_signal_new ("selection_changed",
@@ -200,12 +221,17 @@ eog_wrap_list_instance_init (EogWrapList *wlist)
 	priv->col_spacing = 0;
 	priv->idle_handler_id = -1;
 	priv->n_items = 0;
-	priv->lm = EOG_LAYOUT_MODE_VERTICAL;
 	priv->n_rows = 0;
 	priv->n_cols = 0;
 	priv->last_item_clicked = NULL;
 	priv->item_width = -1;
 	priv->item_height = -1;
+
+	/* Drag source */
+	gtk_drag_source_set (GTK_WIDGET (wlist), 
+			     GDK_BUTTON1_MASK,
+			     drag_types, G_N_ELEMENTS (drag_types),
+			     GDK_ACTION_COPY);
 }
 
 /* Destroy handler for the abstract wrapped list view */
@@ -283,14 +309,12 @@ static gint
 get_item_view_position (EogWrapList *wlist, GnomeCanvasItem *item)
 {
 	EogWrapListPrivate *priv;
-	double x1, y1, x2, y2;
-	gint row, col, n;
+	guint row, col, n;
 	
 	priv = wlist->priv;
-	gnome_canvas_item_get_bounds (item, &x1, &y1, &x2, &y2);
 
-	col = (x1 - priv->col_spacing) / (priv->item_width + priv->col_spacing);
-	row = (y1 - priv->row_spacing) / (priv->item_height + priv->row_spacing);
+	col = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (item), "column"));
+	row = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (item), "row"));
 
 	n = col + priv->n_cols * row;
 
@@ -546,6 +570,18 @@ handle_item_event (GnomeCanvasItem *item, GdkEvent *event,  EogWrapList *wlist)
 	return ret_val;
 }
 
+static void
+handle_item_size_changed (GnomeCanvasItem *item, EogWrapList *wlist)
+{
+	EogWrapListPrivate *priv;
+
+	priv = wlist->priv;
+
+	priv->global_update_hints [ITEM_SIZE_CHANGED] = TRUE;
+
+	request_update (wlist);
+}
+
 /* Size_allocate handler for the abstract wrapped list view */
 static void
 eog_wrap_list_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
@@ -558,7 +594,7 @@ eog_wrap_list_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
 
 	GNOME_CALL_PARENT (GTK_WIDGET_CLASS, size_allocate, (widget, allocation));
 
-	priv->global_update_hints[GLOBAL_SIZE_CHANGED] = TRUE;
+	priv->global_update_hints[GLOBAL_WIDGET_SIZE_CHANGED] = TRUE;
 
 	request_update (wlist);
 }       
@@ -724,6 +760,57 @@ eog_wrap_list_key_press_cb (GtkWidget *widget, GdkEventKey *event)
 	return handled;
 }
 
+static void eog_wrap_list_drag_data_get_cb (GtkWidget *widget,
+					    GdkDragContext *drag_context,
+					    GtkSelectionData *data,
+					    guint info,
+					    guint time)
+{
+	EogWrapList *wlist;
+
+	wlist = EOG_WRAP_LIST (widget);
+
+	switch (info) {
+	case EOG_DND_URI_LIST:
+	{
+		EogImage *image;
+		GList *it;
+		GString *str;
+		GnomeVFSURI *uri;
+
+		/* We assess that every uri has about 30 chars. This
+		   is only to reduce number of reallocations when we
+		   add chars to the string */
+		str = g_string_sized_new (g_list_length (wlist->priv->selected_items) * 30);
+		
+		for (it = wlist->priv->selected_items; it != NULL; it = it->next) {
+			EogCollectionItem *item = EOG_COLLECTION_ITEM (it->data);
+			image = eog_collection_item_get_image (item);
+			uri = eog_image_get_uri (image);
+
+			g_string_append (str, gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_NONE));
+			g_string_append_c (str, '\n');
+
+			gnome_vfs_uri_unref (uri);
+		}
+		
+		gtk_selection_data_set (data,
+					data->target,
+					8, (guchar*) str->str,
+					str->len);
+
+		g_string_free (str, TRUE);
+
+		break;
+	}
+
+	default:
+		g_assert_not_reached ();
+	}
+
+}
+
+
 static gboolean
 add_image (EogCollectionModel *model, EogImage *image, gpointer data)
 {
@@ -738,6 +825,7 @@ add_image (EogCollectionModel *model, EogImage *image, gpointer data)
 	
 	item = eog_collection_item_new (gnome_canvas_root (GNOME_CANVAS (wlist)), image);
 	g_signal_connect (G_OBJECT (item), "event", G_CALLBACK (handle_item_event), wlist);
+	g_signal_connect (G_OBJECT (item), "size_changed", G_CALLBACK (handle_item_size_changed), wlist);
 
 	priv->view_order = g_list_prepend (priv->view_order, item);
 
@@ -763,7 +851,7 @@ model_prepared (EogCollectionModel *model, gpointer data)
 	priv->view_order = g_list_reverse (priv->view_order);
 	priv->n_items = g_list_length (priv->view_order);
 
-	priv->global_update_hints[GLOBAL_SIZE_CHANGED] = TRUE;
+	priv->global_update_hints[GLOBAL_WIDGET_SIZE_CHANGED] = TRUE;
 	request_update (wlist);
 	
 	for (it = priv->view_order; it != NULL; it = it->next) {
@@ -799,7 +887,7 @@ model_image_added (EogCollectionModel *model, EogImage *image, int position, gpo
 	priv->view_order = g_list_insert (priv->view_order, item, position);
 	priv->n_items = g_list_length (priv->view_order);
 
-	priv->global_update_hints[GLOBAL_SIZE_CHANGED] = TRUE;
+	priv->global_update_hints[GLOBAL_WIDGET_SIZE_CHANGED] = TRUE;
 	request_update (wlist);
 	
 	eog_collection_item_load (EOG_COLLECTION_ITEM (item));
@@ -863,23 +951,6 @@ eog_wrap_list_set_model (EogWrapList *wlist, EogCollectionModel *model)
 	request_update (wlist);
 }
 
-void 
-eog_wrap_list_set_layout_mode (EogWrapList *wlist, EogLayoutMode lm)
-{
-	EogWrapListPrivate *priv;
-
-	g_return_if_fail (wlist != NULL);
-	g_return_if_fail (EOG_IS_WRAP_LIST (wlist));
-
-	priv = wlist->priv;
-
-	if (lm == priv->lm) return;
-
-	priv->lm = lm;
-	priv->global_update_hints[GLOBAL_LAYOUT_MODE_CHANGED] = TRUE;
-
-	request_update (wlist);
-}
 
 /**
  * eog_wrap_list_set_row_spacing:
@@ -959,53 +1030,22 @@ do_layout_check (EogWrapList *wlist)
 	g_message ("do_layout_check called");
 #endif
 
-	if (priv->item_width == -1 || priv->item_height == -1) {
-		if (priv->view_order != NULL) {
-			double x1, y1, x2, y2;
-
-			gnome_canvas_item_get_bounds (GNOME_CANVAS_ITEM (priv->view_order->data), &x1, &y1, &x2, &y2);
-			priv->item_width = (int) x2 - x1; 
-			priv->item_height = (int) y2 - y1;
-		}
-		else {
-			return FALSE;
-		}
-	}
-
+	if (priv->view_order == NULL)
+		return FALSE;
 
 	/* get canvas width */
 	cw = GTK_WIDGET (wlist)->allocation.width;
 	ch = GTK_WIDGET (wlist)->allocation.height;
 
 	/* calculate new number of  columns/rows */
-	switch (priv->lm) {
-	case EOG_LAYOUT_MODE_VERTICAL:
-		n_cols_new = cw / (priv->item_width + priv->col_spacing);
-		if (n_cols_new == 0) n_cols_new = 1;
-		n_rows_new = priv->n_items / n_cols_new;
-		n_rows_new = priv->n_items % n_cols_new ? n_rows_new++ : n_rows_new; 
-		break;
-
-	case EOG_LAYOUT_MODE_HORIZONTAL:
-		n_rows_new = ch / (priv->item_height + priv->row_spacing);
-		if (n_rows_new == 0) n_rows_new = 1;
-		n_cols_new = priv->n_items / n_rows_new;
-		n_cols_new = priv->n_items % n_rows_new ? n_cols_new++ : n_cols_new; 
-		break;
-
-	case EOG_LAYOUT_MODE_RECTANGLE:
-		n_rows_new = n_cols_new = sqrt (priv->n_items);
-		if (n_rows_new * n_cols_new < priv->n_items) {
-			if ((n_rows_new+1) * n_cols_new < priv->n_items)
-				n_rows_new = n_cols_new = n_rows_new + 1;
-			else 
-				n_rows_new = n_rows_new + 1;
-		}
-		break;
-
-	default:
-		g_assert_not_reached ();
-	}
+	n_cols_new = cw / (EOG_COLLECTION_ITEM_MAX_WIDTH + priv->col_spacing);
+	
+	if (cw > (n_cols_new * (EOG_COLLECTION_ITEM_MAX_WIDTH + priv->col_spacing) + EOG_COLLECTION_ITEM_MAX_WIDTH))
+		n_cols_new++;
+	
+	if (n_cols_new == 0)
+		n_cols_new = 1;
+	n_rows_new = (priv->n_items + n_cols_new - 1) / n_cols_new;
 
 #if COLLECTION_DEBUG	
 	g_print ("  ** canvas width: %i\n",cw);
@@ -1022,62 +1062,58 @@ do_layout_check (EogWrapList *wlist)
 	return TRUE;
 }
 
-static void 
-calculate_item_position (EogWrapList *wlist, 
-			 guint item_number,
-			 double *world_x,
-			 double *world_y)
-{
-	EogWrapListPrivate *priv;
-	guint row;
-	guint col;
-
-	priv = wlist->priv;
-
-	row = item_number / priv->n_cols;
-	col = item_number % priv->n_cols;
-		
-	*world_x = col * (priv->item_width + priv->col_spacing) + priv->col_spacing;
-	*world_y = row * (priv->item_height + priv->row_spacing) + priv->row_spacing;
-}
-
-typedef struct {
-	EogWrapList *wlist;
-	gint n;
-} RearrangeData; 
-
 static void
-rearrange_single_item (gpointer value, RearrangeData *data)
+set_item_position (EogCollectionItem *item, double x, double y)
 {
-	GnomeCanvasItem *item;
 	double x1, x2, y1, y2;
-	double x3, y3;
 	double xoff, yoff;
+
+	gnome_canvas_item_get_bounds (GNOME_CANVAS_ITEM (item), &x1, &y1, &x2, &y2);
 	
-	item = (GnomeCanvasItem*) value;
-	
-	gnome_canvas_item_get_bounds (item, &x1, &y1, &x2, &y2);
-	
-	calculate_item_position (data->wlist, data->n, &x3, &y3);
-	
-	xoff = x3-x1;
-	yoff = y3-y1;
+	xoff = x - x1;
+	yoff = y - y1;
 		
 	if (xoff || yoff)
-		gnome_canvas_item_move (item, xoff, yoff);
-	
-	data->n++;
+		gnome_canvas_item_move (GNOME_CANVAS_ITEM (item), xoff, yoff);
+}
+
+static void
+calculate_row_height (GList *it, int n_cols, 
+		      int *image_height, int *caption_height)
+{
+	int i;
+	EogCollectionItem *item;
+	int ih, ch, w;
+
+	*image_height = 0;
+	*caption_height = 0;
+
+	for (i = 0; (it != NULL) && (i < n_cols); i++, it = it->next) 
+	{
+		item = EOG_COLLECTION_ITEM (it->data);
+
+		eog_collection_item_get_size (item, &w, &ih, &ch);
+
+		*image_height = MAX (*image_height, ih);
+		*caption_height = MAX (*caption_height, ch);
+	}
 }
 
 static void
 do_item_rearrangement (EogWrapList *wlist)
 {
 	EogWrapListPrivate *priv;
-        RearrangeData data;
+	int max_image_height;
+	int max_caption_height;
+	int item_width;
+	int image_height;
+	int caption_height;
+	int row_offset;
+	int n, row, col;
+	double world_x, world_y;
 	double sr_width, sr_height;
-
-	data.wlist = wlist;
-	data.n = 0;
+	EogCollectionItem *item;
+	GList *it;
 
 #if COLLECTION_DEBUG
 	g_message ("do_item_rearrangement called");
@@ -1085,14 +1121,51 @@ do_item_rearrangement (EogWrapList *wlist)
 
 	priv = wlist->priv;
 
-	g_list_foreach (priv->view_order, 
-			(GFunc) rearrange_single_item, 
-			 &data);
-
+	row_offset = EOG_WRAP_LIST_BORDER - priv->row_spacing;
+	max_image_height = max_caption_height = 0;
+	it = priv->view_order;
+	col = 0;
+	
+	for (n = 0; it != NULL; it = it->next, n++) {
+		col = n % priv->n_cols;
+		row = n / priv->n_cols;
+		
+		if (col == 0) {
+			/* the offset for the new row is the
+			 * old offset plus the height of the
+			 * previous row 
+			 */
+			row_offset = row_offset + max_image_height + max_caption_height + 
+				priv->row_spacing;
+			
+			/* we start a new row and calculate
+			 * the row height/offset for the current row
+			 */
+			calculate_row_height (it, priv->n_cols,
+					      &max_image_height, &max_caption_height);
+		}
+		
+		item = EOG_COLLECTION_ITEM (it->data);
+		eog_collection_item_get_size (item, &item_width, &image_height, &caption_height);
+		item_width = MIN (item_width, MAX_ITEM_WIDTH); /* this is only to ensure 
+								  item_width <= MAX_ITEM_WIDTH */
+		
+		world_x = col * (MAX_ITEM_WIDTH + priv->col_spacing) + 
+			(MAX_ITEM_WIDTH - item_width) / 2;
+		world_y = row_offset + max_image_height - image_height;
+		
+		set_item_position (item, world_x, world_y);
+		
+		g_object_set_data (G_OBJECT (item), "row", GUINT_TO_POINTER (row));
+		g_object_set_data (G_OBJECT (item), "column", GUINT_TO_POINTER (col));
+	}
+	
+	
 	/* set new canvas scroll region */
-	sr_width =  priv->n_cols * (priv->item_width + priv->col_spacing) + priv->col_spacing;
-	sr_height = priv->n_rows * (priv->item_height + priv->row_spacing) + priv->row_spacing;
- 	gnome_canvas_set_scroll_region (GNOME_CANVAS (wlist), 
+	sr_width =  priv->n_cols * (MAX_ITEM_WIDTH + priv->col_spacing) - priv->col_spacing;
+	sr_height = row_offset + max_image_height + max_caption_height + EOG_WRAP_LIST_BORDER;
+	
+	gnome_canvas_set_scroll_region (GNOME_CANVAS (wlist), 
 					0.0, 0.0,
 					sr_width, sr_height);
 
@@ -1115,21 +1188,18 @@ do_update (EogWrapList *wlist)
 	
 	/* handle global updates */
 	if (priv->global_update_hints [GLOBAL_SPACING_CHANGED] ||
-	    priv->global_update_hints [GLOBAL_SIZE_CHANGED]) 
+	    priv->global_update_hints [GLOBAL_WIDGET_SIZE_CHANGED]) 
 	{
 		layout_check_needed = TRUE;
 		item_rearrangement_needed = TRUE;
 		
 		priv->global_update_hints[GLOBAL_SPACING_CHANGED] = FALSE;
-		priv->global_update_hints[GLOBAL_SIZE_CHANGED] = FALSE;
+		priv->global_update_hints[GLOBAL_WIDGET_SIZE_CHANGED] = FALSE;
 	} 
-	else if (priv->global_update_hints [GLOBAL_LAYOUT_MODE_CHANGED]) 
-	{
-		layout_check_needed = TRUE;
-
-		priv->global_update_hints[GLOBAL_LAYOUT_MODE_CHANGED] = FALSE;
+	else if (priv->global_update_hints [ITEM_SIZE_CHANGED]) {
+		item_rearrangement_needed = TRUE;
+		priv->global_update_hints [ITEM_SIZE_CHANGED] = FALSE;
 	}
-
 	
 	if (layout_check_needed && do_layout_check (wlist)) 
 		item_rearrangement_needed = TRUE;
